from canaille.app import models
from canaille.core.populate import fake_groups
from canaille.core.populate import fake_users


def test_delete_group(testclient, backend, user, admin, foo_group):
    foo_group.members = [user, admin]
    backend.save(foo_group)

    user = backend.get(models.User, user.id)
    user.groups = []
    backend.save(user)


def test_no_group(app, backend):
    assert backend.query(models.Group) == []


def test_group_list_pagination(testclient, logged_admin, foo_group, backend):
    res = testclient.get("/groups")
    res.mustcontain("1 item")

    groups = fake_groups(25)

    res = testclient.get("/groups")
    res.mustcontain("26 items")
    group_name = res.pyquery(
        ".groups tbody tr:nth-of-type(1) td:nth-of-type(2) a"
    ).text()
    assert group_name

    form = res.forms["tableform"]
    res = form.submit(name="page", value="2")
    assert group_name not in res.pyquery(
        ".groups tbody tr td:nth-of-type(2) a"
    ).text().split(" ")
    for group in groups:
        backend.delete(group)

    res = testclient.get("/groups")
    res.mustcontain("1 item")


def test_group_list_bad_pages(testclient, logged_admin):
    res = testclient.get("/groups")
    form = res.forms["tableform"]
    testclient.post(
        "/groups", {"csrf_token": form["csrf_token"].value, "page": "2"}, status=404
    )

    res = testclient.get("/groups")
    form = res.forms["tableform"]
    testclient.post(
        "/groups", {"csrf_token": form["csrf_token"].value, "page": "-1"}, status=404
    )


def test_group_deletion(testclient, backend):
    user = models.User(
        formatted_name="foobar",
        family_name="foobar",
        user_name="foobar",
        emails=["foo@bar.test"],
    )
    backend.save(user)

    group = models.Group(
        members=[user],
        display_name="foobar",
    )
    backend.save(group)

    backend.reload(user)
    assert user.groups == [group]

    backend.delete(group)
    backend.reload(user)
    assert not user.groups

    backend.delete(user)


def test_group_list_search(testclient, logged_admin, foo_group, bar_group):
    res = testclient.get("/groups")
    res.mustcontain("2 items")
    res.mustcontain(foo_group.display_name)
    res.mustcontain(bar_group.display_name)

    form = res.forms["search"]
    form["query"] = "oo"
    res = form.submit()

    res.mustcontain("1 item")
    res.mustcontain(foo_group.display_name)
    res.mustcontain(no=bar_group.display_name)


def test_set_groups(app, user, foo_group, bar_group, backend):
    assert user in foo_group.members
    assert user.groups == [foo_group]

    user.groups = [foo_group, bar_group]
    backend.save(user)

    backend.reload(bar_group)
    assert user in bar_group.members
    assert bar_group in user.groups

    user.groups = [foo_group]
    backend.save(user)

    backend.reload(foo_group)
    backend.reload(bar_group)
    assert user in foo_group.members
    assert user not in bar_group.members


def test_set_groups_with_leading_space_in_user_id_attribute(app, foo_group, backend):
    user = models.User(
        formatted_name=" Doe",  # leading space in id attribute
        family_name="Doe",
        user_name="user2",
        emails=["john@doe.test"],
    )
    backend.save(user)

    user.groups = [foo_group]
    backend.save(user)

    backend.reload(foo_group)
    assert user in foo_group.members

    user.groups = []
    backend.save(user)

    backend.reload(foo_group)
    assert user.id not in foo_group.members

    backend.delete(user)


def test_moderator_can_create_edit_and_delete_group(
    testclient, logged_moderator, foo_group, backend
):
    # The group does not exist
    res = testclient.get("/groups", status=200)
    assert backend.get(models.Group, display_name="bar") is None
    assert backend.get(models.Group, display_name="foo") == foo_group
    res.mustcontain(no="bar")
    res.mustcontain("foo")

    # Fill the form for a new group
    res = testclient.get("/groups/add", status=200)
    form = res.forms["creategroupform"]
    form["display_name"] = "bar"
    form["description"] = "yolo"

    # Group has been created
    res = form.submit(status=302).follow(status=200)

    backend.reload(logged_moderator)
    bar_group = backend.get(models.Group, display_name="bar")
    assert bar_group.display_name == "bar"
    assert bar_group.description == "yolo"
    assert bar_group.members == [
        logged_moderator
    ]  # Group cannot be empty so creator is added in it
    res.mustcontain("bar")

    # Group name can not be edited
    res = testclient.get("/groups/bar", status=200)
    form = res.forms["editgroupform"]
    form["display_name"] = "bar2"
    form["description"] = "yolo2"

    res = form.submit(name="action", value="edit")
    assert res.flashes == [("error", "Group edition failed.")]
    res.mustcontain("This field cannot be edited")

    bar_group = backend.get(models.Group, display_name="bar")
    assert bar_group.display_name == "bar"
    assert bar_group.description == "yolo"
    assert backend.get(models.Group, display_name="bar2") is None

    # Group description can be edited
    res = testclient.get("/groups/bar", status=200)
    form = res.forms["editgroupform"]
    form["description"] = "yolo2"

    res = form.submit(name="action", value="edit")
    assert res.flashes == [("success", "The group bar has been successfully edited.")]
    res = res.follow()

    bar_group = backend.get(models.Group, display_name="bar")
    assert bar_group.display_name == "bar"
    assert bar_group.description == "yolo2"

    # Group is deleted
    res = res.forms["editgroupform"].submit(name="action", value="confirm-delete")
    res = res.form.submit(name="action", value="delete", status=302)
    assert backend.get(models.Group, display_name="bar") is None
    assert ("success", "The group bar has been successfully deleted") in res.flashes


def test_cannot_create_already_existing_group(testclient, logged_moderator, foo_group):
    res = testclient.get("/groups/add")
    res = testclient.post(
        "/groups/add",
        {"csrf_token": res.form["csrf_token"].value, "display_name": "foo"},
        status=200,
    )

    res.mustcontain("Group creation failed.")
    res.mustcontain("The group &#39;foo&#39; already exists")


def test_invalid_group(testclient, logged_moderator):
    testclient.get("/groups/invalid", status=404)


def test_simple_user_cannot_view_or_edit_groups(testclient, logged_user, foo_group):
    testclient.get("/groups", status=403)
    testclient.get("/groups/add", status=403)
    testclient.get("/groups/foo", status=403)


def test_invalid_form_request(testclient, logged_moderator, foo_group):
    res = testclient.get("/groups/foo")
    form = res.forms["editgroupform"]
    res = form.submit(name="action", value="invalid-action", status=400)


def test_edition_failed(testclient, logged_moderator, foo_group, backend):
    res = testclient.get("/groups/foo")
    form = res.forms["editgroupform"]
    form["display_name"] = ""
    res = form.submit(name="action", value="edit")
    res.mustcontain("Group edition failed.")
    backend.reload(foo_group)
    assert foo_group.display_name == "foo"


def test_user_list_pagination(testclient, logged_admin, foo_group, backend):
    res = testclient.get("/groups/foo")
    res.mustcontain("1 item")

    users = fake_users(25)
    for user in users:
        foo_group.members = foo_group.members + [user]
    backend.save(foo_group)

    assert len(foo_group.members) == 26
    res = testclient.get("/groups/foo")
    res.mustcontain("26 items")
    user_name = res.pyquery(".users tbody tr:nth-of-type(1) td:nth-of-type(2) a").text()
    assert user_name

    form = res.forms["tableform"]
    res = form.submit(name="page", value="2")
    assert user_name not in res.pyquery(".users tr td:nth-of-type(2) a").text().split(
        " "
    )
    for user in users:
        backend.delete(user)

    res = testclient.get("/groups/foo")
    res.mustcontain("1 item")


def test_user_list_bad_pages(testclient, logged_admin, foo_group):
    res = testclient.get("/groups/foo")
    form = res.forms["tableform"]
    testclient.post(
        "/groups/foo", {"csrf_token": form["csrf_token"].value, "page": "2"}, status=404
    )

    res = testclient.get("/groups/foo")
    form = res.forms["tableform"]
    testclient.post(
        "/groups/foo",
        {"csrf_token": form["csrf_token"].value, "page": "-1"},
        status=404,
    )


def test_user_list_search(
    testclient, logged_admin, foo_group, user, moderator, backend
):
    foo_group.members = foo_group.members + [logged_admin, moderator]
    backend.save(foo_group)

    res = testclient.get("/groups/foo")
    res.mustcontain("3 items")
    res.mustcontain(user.formatted_name)
    res.mustcontain(logged_admin.formatted_name)
    res.mustcontain(moderator.formatted_name)

    form = res.forms["search"]
    form["query"] = "ohn"
    res = form.submit()

    res.mustcontain("1 item")
    res.mustcontain(user.formatted_name)
    res.mustcontain(no=logged_admin.formatted_name)
    res.mustcontain(no=moderator.formatted_name)


def test_remove_member(testclient, logged_admin, foo_group, user, moderator, backend):
    foo_group.members = [user, moderator]
    backend.save(foo_group)

    res = testclient.get("/groups/foo")
    form = res.forms[f"deletegroupmemberform-{user.id}"]

    res = form.submit(name="action", value="confirm-remove-member")
    res = res.form.submit(name="action", value="remove-member")
    assert (
        "success",
        "John (johnny) Doe has been removed from the group foo",
    ) in res.flashes

    backend.reload(foo_group)
    assert user not in foo_group.members


def test_remove_member_already_remove_from_group(
    testclient, logged_admin, foo_group, user, moderator, backend
):
    foo_group.members = [user, moderator]
    backend.save(foo_group)

    res = testclient.get("/groups/foo")
    form = res.forms[f"deletegroupmemberform-{user.id}"]
    foo_group.members = [moderator]
    backend.save(foo_group)

    res = form.submit(name="action", value="confirm-remove-member")
    assert (
        "error",
        "The user 'John (johnny) Doe' has already been removed from the group 'foo'",
    ) in res.flashes


def test_confirm_remove_member_already_removed_from_group(
    testclient, logged_admin, foo_group, user, moderator, backend
):
    foo_group.members = [user, moderator]
    backend.save(foo_group)

    res = testclient.get("/groups/foo")
    form = res.forms[f"deletegroupmemberform-{user.id}"]
    res = form.submit(name="action", value="confirm-remove-member")

    foo_group.members = [moderator]
    backend.save(foo_group)
    res = res.form.submit(name="action", value="remove-member")

    assert (
        "error",
        "The user 'John (johnny) Doe' has already been removed from the group 'foo'",
    ) in res.flashes


def test_remove_member_already_deleted(
    testclient, logged_admin, foo_group, moderator, backend
):
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        user_name="foobar",
        emails=["foobar@example.test"],
    )
    backend.save(user)
    foo_group.members = [user, moderator]
    backend.save(foo_group)

    res = testclient.get("/groups/foo")
    form = res.forms[f"deletegroupmemberform-{user.id}"]
    backend.delete(user)

    res = form.submit(name="action", value="confirm-remove-member")
    assert (
        "error",
        "The user you are trying to remove does not exist.",
    ) in res.flashes


def test_confirm_remove_member_already_deleted(
    testclient, logged_admin, foo_group, moderator, backend
):
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        emails=["foobar@example.test"],
        user_name="foobar",
    )
    backend.save(user)
    foo_group.members = [user, moderator]
    backend.save(foo_group)

    res = testclient.get("/groups/foo")
    form = res.forms[f"deletegroupmemberform-{user.id}"]
    res = form.submit(name="action", value="confirm-remove-member")

    backend.delete(user)
    res = res.form.submit(name="action", value="remove-member")

    assert (
        "error",
        "The user you are trying to remove does not exist.",
    ) in res.flashes



import datetime
import logging

from werkzeug.security import gen_salt

from canaille.app import models


def test_no_logged_no_access(testclient):
    testclient.get("/admin/token", status=403)


def test_no_admin_no_access(testclient, logged_user):
    testclient.get("/admin/token", status=403)


def test_token_list(testclient, token, logged_admin):
    res = testclient.get("/admin/token")
    res.mustcontain(token.token_id)


def test_token_list_pagination(testclient, logged_admin, client, backend):
    res = testclient.get("/admin/token")
    res.mustcontain("0 items")
    tokens = []
    for _ in range(26):
        token = models.Token(
            token_id=gen_salt(48),
            access_token="my-valid-token",
            client=client,
            subject=logged_admin,
            type=None,
            refresh_token=gen_salt(48),
            scope=["openid", "profile"],
            issue_date=(
                datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
            ),
            lifetime=3600,
        )
        backend.save(token)
        tokens.append(token)

    res = testclient.get("/admin/token")
    res.mustcontain("26 items")
    token_id = res.pyquery(".tokens tbody tr td:nth-of-type(1) a").text()
    assert token_id

    form = res.forms["tableform"]
    res = form.submit(name="form", value="2")
    assert token_id not in res.pyquery(
        ".tokens tbody tr:nth-of-type(1) td:nth-of-type(1) a"
    ).text().split(" ")
    for token in tokens:
        backend.delete(token)

    res = testclient.get("/admin/token")
    res.mustcontain("0 items")


def test_token_list_bad_pages(testclient, logged_admin):
    res = testclient.get("/admin/token")
    form = res.forms["tableform"]
    testclient.post(
        "/admin/token",
        {"csrf_token": form["csrf_token"].value, "page": "2"},
        status=404,
    )

    res = testclient.get("/admin/token")
    form = res.forms["tableform"]
    testclient.post(
        "/admin/token",
        {"csrf_token": form["csrf_token"].value, "page": "-1"},
        status=404,
    )


def test_token_list_search(testclient, logged_admin, client, backend):
    token1 = models.Token(
        token_id=gen_salt(48),
        access_token="this-token-is-ok",
        client=client,
        subject=logged_admin,
        type=None,
        refresh_token=gen_salt(48),
        scope=["openid", "profile"],
        issue_date=(
            datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        ),
        lifetime=3600,
    )
    backend.save(token1)
    token2 = models.Token(
        token_id=gen_salt(48),
        access_token="this-token-is-valid",
        client=client,
        subject=logged_admin,
        type=None,
        refresh_token=gen_salt(48),
        scope=["openid", "profile"],
        issue_date=(
            datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        ),
        lifetime=3600,
    )
    backend.save(token2)

    res = testclient.get("/admin/token")
    res.mustcontain("2 items")
    res.mustcontain(token1.token_id)
    res.mustcontain(token2.token_id)

    form = res.forms["search"]
    form["query"] = "valid"
    res = form.submit()

    res.mustcontain("1 item")
    res.mustcontain(token2.token_id)
    res.mustcontain(no=token1.token_id)


def test_token_view(testclient, token, logged_admin):
    res = testclient.get("/admin/token/" + token.token_id)
    res.mustcontain(token.access_token)


def test_token_not_found(testclient, logged_admin):
    testclient.get("/admin/token/" + "yolo", status=404)


def test_revoke_bad_request(testclient, token, logged_admin):
    assert not token.revoked

    res = testclient.get(f"/admin/token/{token.token_id}")
    res = res.form.submit(name="action", value="invalid", status=400)


def test_revoke_token(testclient, token, logged_admin, backend, caplog):
    assert not token.revoked

    res = testclient.get(f"/admin/token/{token.token_id}")
    res = res.form.submit(name="action", value="confirm-revoke")
    res = res.form.submit(name="action", value="revoke")
    assert ("success", "The token has successfully been revoked.") in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Revoked token for user in client Some client by admin",
    ) in caplog.record_tuples

    backend.reload(token)
    assert token.revoked


def test_revoke_invalid_token(testclient, logged_admin):
    testclient.get("/admin/token/invalid/revoke", status=404)



import json
from unittest import mock

from canaille.commands import cli


def test_dump_stdout(testclient, backend, user, foo_group):
    """Test the full database dump command."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["dump"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "authorizationcode": [],
        "client": [],
        "consent": [],
        "group": [
            {
                "created": mock.ANY,
                "display_name": "foo",
                "id": foo_group.id,
                "last_modified": mock.ANY,
                "members": [
                    user.id,
                ],
            },
        ],
        "token": [],
        "user": [
            {
                "created": mock.ANY,
                "display_name": "Johnny",
                "emails": [
                    "john@doe.test",
                ],
                "family_name": "Doe",
                "formatted_address": "1235, somewhere",
                "formatted_name": "John (johnny) Doe",
                "given_name": "John",
                "groups": [foo_group.id],
                "id": user.id,
                "last_modified": mock.ANY,
                "password": mock.ANY,
                "phone_numbers": [
                    "555-000-000",
                ],
                "preferred_language": "en",
                "profile_url": "https://john.test",
                "user_name": "user",
            },
        ],
    }



from flask import current_app
from flask import flash

from canaille.app.i18n import gettext as _
from canaille.core.mails import send_compromised_password_check_failure_mail


def check_if_send_mail_to_admins(form, api_url, hashed_password_suffix):
    if (
        not current_app.features.has_smtp
        or not current_app.config["CANAILLE"]["ADMIN_EMAIL"]
    ):
        return

    flash(
        _(
            "Password compromise investigation failed. "
            "Please contact the administrators."
        ),
        "error",
    )

    if form.user is not None:
        user_name = form.user.user_name
        user_email = form.user.emails[0]
    else:
        user_name = form["user_name"].data
        user_email = form["emails"].data[0]

    if send_compromised_password_check_failure_mail(
        api_url,
        user_name,
        user_email,
        hashed_password_suffix,
        current_app.config["CANAILLE"]["ADMIN_EMAIL"],
    ):
        flash(
            _(
                "We have informed your administrator about the failure of the password compromise investigation."
            ),
            "info",
        )
    else:
        flash(
            _(
                "An error occurred while communicating the incident to the administrators. "
                "Please update your password as soon as possible. "
                "If this still happens, please contact the administrators."
            ),
            "error",
        )



from flask import current_app


def send_sms(recipient, sender, text):
    try:
        import smpplib.client
        import smpplib.consts
        import smpplib.gsm
    except ImportError as exc:
        raise RuntimeError(
            "You are trying to send a sms but the 'sms' extra is not installed."
        ) from exc

    port = current_app.config["CANAILLE"]["SMPP"]["PORT"]
    host = current_app.config["CANAILLE"]["SMPP"]["HOST"]
    login = current_app.config["CANAILLE"]["SMPP"]["LOGIN"]
    password = current_app.config["CANAILLE"]["SMPP"]["PASSWORD"]

    try:
        client = smpplib.client.Client(host, port, allow_unknown_opt_params=True)
        client.connect()
        try:
            client.bind_transmitter(system_id=login, password=password)
            pdu = client.send_message(
                source_addr_ton=smpplib.consts.SMPP_TON_INTL,
                source_addr=sender,
                dest_addr_ton=smpplib.consts.SMPP_TON_INTL,
                destination_addr=recipient,
                short_message=bytes(text, "utf-8"),
            )
            current_app.logger.debug(pdu.generate())
        finally:
            if client.state in [
                smpplib.consts.SMPP_CLIENT_STATE_BOUND_TX
            ]:  # pragma: no cover
                # if bound to transmitter
                try:
                    client.unbind()
                except smpplib.exceptions.UnknownCommandError:
                    try:
                        client.unbind()
                    except smpplib.exceptions.PDUError:
                        pass
    except Exception as exc:
        current_app.logger.warning(f"Could not send sms: {exc}")
        return False
    finally:
        if client:  # pragma: no branch
            client.disconnect()

    return True



from pydantic import Field

from canaille.app.configuration import BaseModel


class LDAPSettings(BaseModel):
    """Settings related to the LDAP backend.

    Belong in the ``CANAILLE_LDAP`` namespace.
    """

    URI: str = Field(
        "ldap://localhost", examples=["ldap://localhost", "ldaps://ldap.example.org"]
    )
    """The LDAP server URI."""

    ROOT_DN: str = "dc=example,dc=org"
    """The LDAP root DN."""

    BIND_DN: str = "cn=admin,dc=example,dc=org"
    """The LDAP bind DN."""

    BIND_PW: str = "admin"
    """The LDAP bind password."""

    TIMEOUT: float = 0.0
    """The LDAP connection timeout."""

    USER_BASE: str = Field(..., examples=["ou=users,dc=example,dc=org"])
    """The LDAP node under which users will be looked for and saved.

    For instance `ou=users,dc=example,dc=org`.
    """

    USER_CLASS: list[str] = ["inetOrgPerson"]
    """The object class to use for creating new users."""

    USER_RDN: str = "uid"
    """The attribute to identify an object in the User DN."""

    USER_FILTER: str = "(|(uid={{ login }})(mail={{ login }}))"
    """Filter to match users on sign in.

    For instance ``(|(uid={{ login }})(mail={{ login }}))``.
    Jinja syntax is supported and a ``login`` variable is available,
    containing the value passed in the login field.
    """

    GROUP_BASE: str = Field(..., examples=["ou=groups,dc=example,dc=org"])
    """The LDAP node under which groups will be looked for and saved.

    For instance `"ou=groups,dc=example,dc=org"`.
    """

    GROUP_CLASS: str = "groupOfNames"
    """The object class to use for creating new groups."""

    GROUP_RDN: str = "cn"
    """The attribute to identify an object in the Group DN."""

    GROUP_NAME_ATTRIBUTE: str = "cn"
    """The attribute to use to identify a group."""



#!/usr/bin/env python3
import datetime
import os
import pathlib
import sys
from importlib import metadata
from unittest import mock

sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("../canaille"))

from canaille.app.configuration import export_config
from canaille.app.configuration import settings_factory

# Readthedocs does not support C modules, so
# we have to mock them.


class Mock(mock.MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return mock.MagicMock()


MOCK_MODULES = ["ldap"]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

# -- General configuration ------------------------------------------------


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_click",
    "sphinx_design",
    "sphinx_issues",
    "sphinxcontrib.autodoc_pydantic",
    "jinja_autodoc",
    "sphinxcontrib.screenshot",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}
master_doc = "index"
project = "canaille"
year = datetime.datetime.now().strftime("%Y")
copyright = f"{year}, Yaal Coop"
author = "Yaal Coop"

version = metadata.version("canaille")
language = "en"
exclude_patterns = []
pygments_style = "sphinx"
todo_include_todos = True
toctree_collapse = False
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "authlib": ("https://docs.authlib.org/en/latest", None),
    "flask": ("https://flask.palletsprojects.com", None),
    "flask-alembic": ("https://flask-alembic.readthedocs.io/en/latest", None),
    "flask-babel": ("https://python-babel.github.io/flask-babel", None),
    "flask-wtf": ("https://flask-wtf.readthedocs.io", None),
    "hypercorn": ("https://hypercorn.readthedocs.io/en/latest", None),
    "pydantic": ("https://docs.pydantic.dev/latest", None),
    "pytest-iam": ("https://pytest-iam.readthedocs.io/en/latest/", None),
    "wtforms": ("https://wtforms.readthedocs.io", None),
    "scim2-cli": ("https://scim2-cli.readthedocs.io/en/latest", None),
}

issues_uri = "https://gitlab.com/yaal/canaille/-/issues/{issue}"
issues_pr_uri = "https://gitlab.com/yaal/canaille/-/merge_requests/{pr}"
issues_commit_uri = "https://gitlab.com/yaal/canaille/-/commit/{commit}"

# -- Options for HTML output ----------------------------------------------

html_theme = "shibuya"
html_static_path = ["_static"]
html_baseurl = "https://canaille.readthedocs.io/"
html_theme_options = {
    "globaltoc_expand_depth": 3,
    "accent_color": "yellow",
    "light_logo": "_static/canaille-label-black.webp",
    "dark_logo": "_static/canaille-label-white.webp",
    "gitlab_url": "https://gitlab.com/yaal/canaille",
    "mastodon_url": "https://toot.aquilenet.fr/@yaal",
    "discussion_url": "https://matrix.to/#/#canaille-discuss:yaal.coop",
    "nav_links": [
        {
            "title": "Demo",
            "children": [
                {
                    "title": "Canaille demo server",
                    "url": "https://demo.canaille.yaal.coop",
                },
                {
                    "title": "OIDC Client 1",
                    "url": "https://demo.client1.yaal.coop",
                },
                {
                    "title": "OIDC Client 2",
                    "url": "https://demo.client2.yaal.coop",
                },
            ],
        },
        {"title": "PyPI", "url": "https://pypi.org/project/Canaille"},
        {
            "title": "Weblate",
            "url": "https://hosted.weblate.org/projects/canaille/canaille",
        },
    ],
}
html_context = {
    "source_type": "gitlab",
    "source_user": "yaal",
    "source_repo": "canaille",
    "source_version": "main",
    "source_docs_path": "/doc/",
    "languages": [
        ("English", "/en/latest/%s.html"),
        ("Français", "/fr/latest/%s.html"),
    ],
}

# -- Options for HTMLHelp output ------------------------------------------

htmlhelp_basename = "canailledoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}
latex_documents = [
    (master_doc, "canaille.tex", "canaille Documentation", "Yaal", "manual")
]

# -- Options for manual page output ---------------------------------------

man_pages = [(master_doc, "canaille", "canaille Documentation", [author], 1)]

# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (
        master_doc,
        "canaille",
        "canaille Documentation",
        author,
        "canaille",
        "One line description of project.",
        "Miscellaneous",
    )
]

# -- Options for autosectionlabel -----------------------------------------

autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

# -- Options for autodoc_pydantic_settings -------------------------------------------

autodoc_pydantic_settings_show_json = False
autodoc_pydantic_settings_show_config_summary = False
autodoc_pydantic_settings_show_config_summary = False
autodoc_pydantic_settings_show_validator_summary = False
autodoc_pydantic_settings_show_validator_members = False
autodoc_pydantic_settings_show_field_summary = False
autodoc_pydantic_settings_signature_prefix = ""
autodoc_pydantic_field_signature_prefix = ""
autodoc_pydantic_field_list_validators = False
autodoc_pydantic_field_doc_policy = "docstring"

# -- Translation options ------------------------------------------------------
# Advised by https://docs.readthedocs.io/en/latest/guides/manage-translations-sphinx.html#create-translatable-files
gettext_uuid = True
gettext_compact = "doc"

# -- Templates options ------------------------------------------------------
jinja_template_path = str(
    pathlib.Path(__file__).parent.parent.resolve() / "canaille/templates"
)

# -- Screenshots options ------------------------------------------------------
screenshot_apps = {
    "canaille": "doc.doc_app:create_doc_app",
}

screenshot_contexts = {
    "admin": "doc.doc_app:admin_login",
    "user": "doc.doc_app:user_login",
    "james": "doc.doc_app:james_login",
}
screenshot_default_browser = "firefox"
screenshot_default_viewport_width = 770
screenshot_default_viewport_height = 1
screenshot_default_full_page = True


config_obj = settings_factory(all_options=True, init_with_examples=True)
export_config(config_obj, "config.sample.toml")



from flask import g


def test_index(testclient, user, backend):
    res = testclient.get("/", status=302)
    assert res.location == "/login"

    g.user = user
    res = testclient.get("/", status=302)
    assert res.location == "/profile/user"

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = ["use_oidc"]
    backend.reload(g.user)
    res = testclient.get("/", status=302)
    assert res.location == "/consent/"

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = []
    backend.reload(g.user)
    res = testclient.get("/", status=302)
    assert res.location == "/about"



import datetime
import uuid

from flask import Blueprint
from flask import current_app
from flask import flash
from flask import redirect
from flask import url_for

from canaille.app import models
from canaille.app.flask import user_needed
from canaille.app.i18n import gettext as _
from canaille.app.templating import render_template
from canaille.backends import Backend

from ..utils import SCOPE_DETAILS

bp = Blueprint("consents", __name__, url_prefix="/consent")


@bp.route("/")
@user_needed()
def consents(user):
    consents = Backend.instance.query(models.Consent, subject=user)
    clients = {t.client for t in consents}

    nb_consents = len(consents)
    nb_preconsents = sum(
        1
        for client in Backend.instance.query(models.Client)
        if client.preconsent and client not in clients
    )

    return render_template(
        "oidc/consent_list.html",
        consents=consents,
        menuitem="consents",
        scope_details=SCOPE_DETAILS,
        ignored_scopes=["openid"],
        nb_consents=nb_consents,
        nb_preconsents=nb_preconsents,
    )


@bp.route("/pre-consents")
@user_needed()
def pre_consents(user):
    consents = Backend.instance.query(models.Consent, subject=user)
    clients = {t.client for t in consents}
    preconsented = [
        client
        for client in Backend.instance.query(models.Client)
        if client.preconsent and client not in clients
    ]

    nb_consents = len(consents)
    nb_preconsents = len(preconsented)

    return render_template(
        "oidc/preconsent_list.html",
        menuitem="consents",
        scope_details=SCOPE_DETAILS,
        # TODO: do not delegate this var to the templates, or set this explicitly in the templates.
        ignored_scopes=["openid"],
        preconsented=preconsented,
        nb_consents=nb_consents,
        nb_preconsents=nb_preconsents,
    )


@bp.route("/revoke/<consent(required=False):consent>")
@user_needed()
def revoke(user, consent):
    if not consent or consent.subject != user:
        flash(_("Could not revoke this access"), "error")

    elif consent.revokation_date:
        flash(_("The access is already revoked"), "error")

    else:
        consent.revoke()
        current_app.logger.security(
            f"Consent revoked for {user.user_name} in client {consent.client.client_name}"
        )
        flash(_("The access has been revoked"), "success")

    return redirect(url_for("oidc.consents.consents"))


@bp.route("/restore/<consent(required=False):consent>")
@user_needed()
def restore(user, consent):
    if not consent or consent.subject != user:
        flash(_("Could not restore this access"), "error")

    elif not consent.revokation_date:
        flash(_("The access is not revoked"), "error")

    else:
        consent.restore()
        if not consent.issue_date:
            consent.issue_date = datetime.datetime.now(datetime.timezone.utc)
        Backend.instance.save(consent)
        flash(_("The access has been restored"), "success")

    return redirect(url_for("oidc.consents.consents"))


@bp.route("/revoke-preconsent/<client(required=False):client>")
@user_needed()
def revoke_preconsent(user, client):
    if not client or not client.preconsent:
        flash(_("Could not revoke this access"), "error")
        return redirect(url_for("oidc.consents.consents"))

    consent = Backend.instance.get(models.Consent, client=client, subject=user)
    if consent:
        return redirect(url_for("oidc.consents.revoke", consent=consent))

    consent = models.Consent(
        consent_id=str(uuid.uuid4()),
        client=client,
        subject=user,
        scope=client.scope,
    )
    consent.revoke()
    Backend.instance.save(consent)
    flash(_("The access has been revoked"), "success")

    return redirect(url_for("oidc.consents.consents"))



import pytest

from canaille.backends.memory.backend import MemoryBackend


@pytest.fixture
def memory_backend(configuration):
    backend = MemoryBackend(configuration)
    with backend.session():
        yield backend



def test_version(testclient):
    testclient.get("/about", status=200)


def test_javascript(testclient):
    testclient.app.config["CANAILLE"]["JAVASCRIPT"] = True
    res = testclient.get("/about", status=200)
    res.mustcontain("<script>")

    testclient.app.config["CANAILLE"]["JAVASCRIPT"] = False
    res = testclient.get("/about", status=200)
    res.mustcontain(no="<script>")


def test_htmx(testclient):
    testclient.app.config["CANAILLE"]["JAVASCRIPT"] = True
    testclient.app.config["CANAILLE"]["HTMX"] = True
    res = testclient.get("/about", status=200)
    res.mustcontain("htmx.min.js")
    res.mustcontain("hx-boost")

    testclient.app.config["CANAILLE"]["JAVASCRIPT"] = True
    testclient.app.config["CANAILLE"]["HTMX"] = False
    res = testclient.get("/about", status=200)
    res.mustcontain(no="htmx.min.js")
    res.mustcontain(no="hx-boost")



import logging
from logging.config import dictConfig
from logging.config import fileConfig

from flask import has_request_context
from flask import request
from flask.logging import default_handler
from flask.logging import wsgi_errors_stream


class IPFilter(logging.Filter):
    def filter(self, record):
        record.ip = request.remote_addr if has_request_context() else ""
        return True


def add_log_level(level_name, level_num, method_name=None):
    """Adapted from https://stackoverflow.com/a/35804945/2700168.

    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.
    `level_name` becomes an attribute of the `logging` module with the value
    `level_num`.
    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example:
    -------
    >>> addLoggingLevel("TRACE", logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace("that worked")
    >>> logging.trace("so did this")
    >>> logging.TRACE
    5

    """
    if not method_name:
        method_name = level_name.lower()

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def setup_logging(app):
    conf = app.config["CANAILLE"]["LOGGING"]

    security_level_name = "SECURITY"
    # SECURITY is between INFO and WARNING
    security_level = logging.INFO + 5

    if not hasattr(logging, security_level_name):
        add_log_level(security_level_name, security_level)

    if conf is None:
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.WARNING)

        log_level = "DEBUG" if app.debug else "INFO"
        formatter = logging.Formatter(
            "[%(asctime)s] - %(ip)s - %(levelname)s in %(module)s: %(message)s"
        )
        handler = logging.StreamHandler(stream=wsgi_errors_stream)
        handler.setFormatter(formatter)
        app.logger.setLevel(log_level)
        app.logger.removeHandler(default_handler)
        app.logger.addHandler(handler)

    elif isinstance(conf, dict):
        dictConfig(conf)

    else:
        fileConfig(conf, disable_existing_loggers=False)

    app.logger.addFilter(IPFilter())



from canaille.app import models
from canaille.commands import cli


def test_delete_by_id(testclient, backend):
    """Remove a model identified by its id."""
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        emails=["foobar@example.test"],
        user_name="foobar",
    )
    backend.save(user)

    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["delete", "user", user.id])
    assert res.exit_code == 0, res.stdout

    assert not backend.get(models.User, user_name="foobar")


def test_delete_by_identifier(testclient, backend):
    """Remove a model identified by its identifier."""
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        emails=["foobar@example.test"],
        user_name="foobar",
    )
    backend.save(user)

    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["delete", "user", "foobar"])
    assert res.exit_code == 0, res.stdout

    assert not backend.get(models.User, user_name="foobar")


def test_delete_unknown_id(testclient, backend):
    """Error case for trying to set a value for an invalid object."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(cli, ["delete", "user", "invalid"])
    assert res.exit_code == 1, res.stdout
    assert res.stdout == "Error: No user with id 'invalid'\n"



import datetime
import logging

import pytest
import time_machine

from canaille.app import models
from canaille.core.models import HOTP_LOOK_AHEAD_WINDOW


def test_otp_disabled(testclient):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = None
    with testclient.session_transaction() as session:
        session["remaining_otp_methods"] = ["TOTP"]
        session["attempt_login_with_correct_password"] = "id"
    testclient.get("/setup-mfa", status=404)
    testclient.get("/verify-mfa", status=404)


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_signin_and_out_with_otp(testclient, user_otp, caplog, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "user" == session.get("attempt_login")

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)
    assert (
        "info",
        "Please enter the one-time password from your authenticator app.",
    ) in res.flashes
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "attempt_login" not in session
        assert "user" == session.get("attempt_login_with_correct_password")

    res = testclient.get("/verify-mfa")
    res.form["otp"] = user_otp.generate_otp()
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome John (johnny) Doe",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Succeed login attempt for user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert [user_otp.id] == session.get("user_id")
        assert "attempt_login" not in session
        assert "attempt_login_with_correct_password" not in session

    res = testclient.get("/login", status=302)

    res = testclient.get("/logout")
    assert (
        "success",
        "You have been disconnected. See you next time John (johnny) Doe",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Logout user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_signin_wrong_otp(testclient, user_otp, caplog, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["otp"] = 111111
    res = res.form.submit()

    assert (
        "error",
        "The one-time password you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed login attempt (wrong OTP) for user",
    ) in caplog.record_tuples


def test_signin_expired_totp(testclient, user_otp, caplog):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "TOTP"

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False) as traveller:
        with testclient.session_transaction() as session:
            assert not session.get("user_id")

        res = testclient.get("/login", status=200)

        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)

        res.form["password"] = "correct horse battery staple"
        res = res.form.submit(status=302)
        res = res.follow(status=200)

        res.form["otp"] = user_otp.generate_otp()
        traveller.shift(datetime.timedelta(seconds=30))
        res = res.form.submit()

        assert (
            "error",
            "The one-time password you entered is invalid. Please try again",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Failed login attempt (wrong OTP) for user",
        ) in caplog.record_tuples


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_new_user_setup_otp(testclient, backend, caplog, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    u = models.User(
        formatted_name="Otp User",
        family_name="Otp",
        user_name="otp",
        emails=["john@doe.test"],
        password="correct horse battery staple",
    )
    backend.save(u)

    assert u.secret_token is not None
    if otp_method == "HOTP":
        assert u.hotp_counter == 1

    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "otp"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)

    assert res.location == "/setup-mfa"
    assert (
        "info",
        "You have not enabled multi-factor authentication. Please enable it first to login.",
    ) in res.flashes
    res = testclient.get("/setup-mfa", status=200)
    assert u.secret_token == res.form["secret"].value

    res = testclient.get("/verify-mfa", status=200)
    res.form["otp"] = u.generate_otp()
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome Otp User",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Succeed login attempt for otp",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)
    with testclient.session_transaction() as session:
        assert [u.id] == session.get("user_id")
        assert "attempt_login" not in session
        assert "attempt_login_with_correct_password" not in session

    res = testclient.get("/login", status=302)

    backend.delete(u)


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_verify_otp_page_without_signin_in_redirects_to_login_page(
    testclient, user_otp, otp_method
):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    res = testclient.get("/verify-mfa", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with")
    ]


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_setup_otp_page_without_signin_in_redirects_to_login_page(
    testclient, user_otp, otp_method
):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    res = testclient.get("/setup-mfa", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with")
    ]


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_verify_otp_page_already_logged_in(testclient, logged_user_otp, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    res = testclient.get("/verify-mfa", status=302)
    assert res.location == "/profile/user"


def test_signin_multiple_attempts_doesnt_desynchronize_hotp(
    testclient, user_otp, caplog
):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "HOTP"

    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "user" == session.get("attempt_login")

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "attempt_login" not in session
        assert "user" == session.get("attempt_login_with_correct_password")

    res = testclient.get("/verify-mfa")
    for _x in range(3):
        res.form["otp"] = "111111"
        res = res.form.submit(status=302).follow()
    res.form["otp"] = user_otp.generate_otp()
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome John (johnny) Doe",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Succeed login attempt for user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_setup_otp_page_already_logged_in(testclient, logged_user_otp, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    res = testclient.get("/setup-mfa", status=302)
    assert res.location == "/profile/user"


def test_signin_inside_hotp_look_ahead_window(testclient, backend, user_otp, caplog):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "HOTP"

    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    assert user_otp.hotp_counter == 1

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "user" == session.get("attempt_login")

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "attempt_login" not in session
        assert "user" == session.get("attempt_login_with_correct_password")

    res = testclient.get("/verify-mfa")

    res.form["otp"] = user_otp.generate_otp(HOTP_LOOK_AHEAD_WINDOW)
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome John (johnny) Doe",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Succeed login attempt for user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)

    user = backend.get(models.User, id=user_otp.id)
    assert user.hotp_counter == HOTP_LOOK_AHEAD_WINDOW + 2


def test_signin_outside_hotp_look_ahead_window(testclient, backend, user_otp, caplog):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "HOTP"

    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    assert user_otp.hotp_counter == 1

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "user" == session.get("attempt_login")

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "attempt_login" not in session
        assert "user" == session.get("attempt_login_with_correct_password")

    res = testclient.get("/verify-mfa")

    res.form["otp"] = user_otp.generate_otp(HOTP_LOOK_AHEAD_WINDOW + 1)
    res = res.form.submit(status=302)

    assert (
        "error",
        "The one-time password you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed login attempt (wrong OTP) for user",
    ) in caplog.record_tuples

    user = backend.get(models.User, id=user_otp.id)
    assert user.hotp_counter == 1



import datetime

from authlib.integrations.flask_oauth2 import AuthorizationServer
from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oauth2.rfc6749.grants import ClientCredentialsGrant
from authlib.oauth2.rfc6749.grants import ImplicitGrant
from authlib.oauth2.rfc6749.grants import RefreshTokenGrant as _RefreshTokenGrant
from authlib.oauth2.rfc6749.grants import (
    ResourceOwnerPasswordCredentialsGrant as _ResourceOwnerPasswordCredentialsGrant,
)
from authlib.oauth2.rfc6750 import BearerTokenValidator as _BearerTokenValidator
from authlib.oauth2.rfc7009 import RevocationEndpoint as _RevocationEndpoint
from authlib.oauth2.rfc7591 import (
    ClientRegistrationEndpoint as _ClientRegistrationEndpoint,
)
from authlib.oauth2.rfc7592 import (
    ClientConfigurationEndpoint as _ClientConfigurationEndpoint,
)
from authlib.oauth2.rfc7636 import CodeChallenge as _CodeChallenge
from authlib.oauth2.rfc7662 import IntrospectionEndpoint as _IntrospectionEndpoint
from authlib.oidc.core import UserInfo
from authlib.oidc.core.grants import OpenIDCode as _OpenIDCode
from authlib.oidc.core.grants import OpenIDHybridGrant as _OpenIDHybridGrant
from authlib.oidc.core.grants import OpenIDImplicitGrant as _OpenIDImplicitGrant
from authlib.oidc.core.grants.util import generate_id_token
from flask import current_app
from flask import g
from flask import request
from flask import url_for
from joserfc.jwk import JWKRegistry
from werkzeug.security import gen_salt

from canaille.app import DOCUMENTATION_URL
from canaille.app import models
from canaille.backends import Backend

AUTHORIZATION_CODE_LIFETIME = 84400


def oauth_authorization_server():
    return {
        "issuer": get_issuer(),
        "authorization_endpoint": url_for("oidc.endpoints.authorize", _external=True),
        "token_endpoint": url_for("oidc.endpoints.issue_token", _external=True),
        "token_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "private_key_jwt",
            "client_secret_post",
            "none",
        ],
        "token_endpoint_auth_signing_alg_values_supported": ["RS256", "ES256"],
        "userinfo_endpoint": url_for("oidc.endpoints.userinfo", _external=True),
        "revocation_endpoint": url_for("oidc.endpoints.revoke_token", _external=True),
        "introspection_endpoint": url_for(
            "oidc.endpoints.introspect_token", _external=True
        ),
        "jwks_uri": url_for("oidc.endpoints.jwks", _external=True),
        "registration_endpoint": url_for(
            "oidc.endpoints.client_registration", _external=True
        ),
        "scopes_supported": [
            "openid",
            "profile",
            "email",
            "address",
            "phone",
            "groups",
        ],
        "response_types_supported": [
            "code",
            "token",
            "id_token",
            "code token",
            "code id_token",
            "token id_token",
        ],
        "grant_types": [
            "authorization_code",
            "implicit",
            "password",
            "client_credentials",
            "refresh_token",
        ],
        "ui_locales_supported": g.available_language_codes,
        "code_challenge_methods_supported": ["plain", "S256"],
        "service_documentation": DOCUMENTATION_URL,
    }


def openid_configuration():
    return {
        **oauth_authorization_server(),
        "end_session_endpoint": url_for("oidc.endpoints.end_session", _external=True),
        "claims_supported": [
            "sub",
            "iss",
            "auth_time",
            "acr",
            "name",
            "given_name",
            "family_name",
            "nickname",
            "profile",
            "picture",
            "website",
            "email",
            "email_verified",
            "locale",
            "zoneinfo",
            "groups",
            "nonce",
        ],
        "subject_types_supported": ["pairwise", "public"],
        "id_token_signing_alg_values_supported": ["RS256", "ES256", "HS256"],
        "prompt_values_supported": ["none"]
        + (["create"] if current_app.config["CANAILLE"]["ENABLE_REGISTRATION"] else []),
    }


def exists_nonce(nonce, req):
    client = Backend.instance.get(models.Client, id=req.client_id)
    exists = Backend.instance.query(
        models.AuthorizationCode, client=client, nonce=nonce
    )
    return bool(exists)


def get_issuer():
    if current_app.config["CANAILLE_OIDC"]["JWT"]["ISS"]:
        return current_app.config["CANAILLE_OIDC"]["JWT"]["ISS"]

    if current_app.config.get("SERVER_NAME"):
        return current_app.config.get("SERVER_NAME")

    return request.url_root


def get_jwt_config(grant=None):
    return {
        "key": current_app.config["CANAILLE_OIDC"]["JWT"]["PRIVATE_KEY"],
        "alg": current_app.config["CANAILLE_OIDC"]["JWT"]["ALG"],
        "iss": get_issuer(),
        "exp": current_app.config["CANAILLE_OIDC"]["JWT"]["EXP"],
    }


def get_jwks():
    kty = current_app.config["CANAILLE_OIDC"]["JWT"]["KTY"]
    alg = current_app.config["CANAILLE_OIDC"]["JWT"]["ALG"]
    jwk = JWKRegistry.import_key(
        current_app.config["CANAILLE_OIDC"]["JWT"]["PUBLIC_KEY"], kty
    )
    jwk.ensure_kid()
    return {
        "keys": [
            {
                "kid": jwk.kid,
                "use": "sig",
                "alg": alg,
                **jwk,
            }
        ]
    }


def claims_from_scope(scope):
    claims = {"sub"}
    if "profile" in scope:
        claims |= {
            "name",
            "family_name",
            "given_name",
            "nickname",
            "preferred_username",
            "profile",
            "picture",
            "website",
            "gender",
            "birthdate",
            "zoneinfo",
            "locale",
            "updated_at",
        }
    if "email" in scope:
        claims |= {"email", "email_verified"}
    if "address" in scope:
        claims |= {"address"}
    if "phone" in scope:
        claims |= {"phone_number", "phone_number_verified"}
    if "groups" in scope:
        claims |= {"groups"}
    return claims


def generate_user_info(user, scope):
    claims = claims_from_scope(scope)
    data = generate_user_claims(user, claims)
    return UserInfo(**data)


def generate_user_claims(user, claims, jwt_mapping_config=None):
    jwt_mapping_config = {
        **(current_app.config["CANAILLE_OIDC"]["JWT"]["MAPPING"]),
        **(jwt_mapping_config or {}),
    }

    data = {}
    for claim in claims:
        raw_claim = jwt_mapping_config.get(claim.upper())
        if raw_claim:
            formatted_claim = current_app.jinja_env.from_string(raw_claim).render(
                user=user
            )
            if formatted_claim:
                # According to https://openid.net/specs/openid-connect-core-1_0.html#UserInfoResponse
                # it's better to not insert a null or empty string value
                data[claim] = formatted_claim
        if claim == "groups":
            data[claim] = [group.display_name for group in user.groups]
    return data


def save_authorization_code(code, request):
    nonce = request.data.get("nonce")
    now = datetime.datetime.now(datetime.timezone.utc)
    scope = request.client.get_allowed_scope(request.scope)
    code = models.AuthorizationCode(
        authorization_code_id=gen_salt(48),
        code=code,
        subject=request.user,
        client=request.client,
        redirect_uri=request.redirect_uri or request.client.redirect_uris[0],
        scope=scope.split(" "),
        nonce=nonce,
        issue_date=now,
        lifetime=AUTHORIZATION_CODE_LIFETIME,
        challenge=request.data.get("code_challenge"),
        challenge_method=request.data.get("code_challenge_method"),
    )
    Backend.instance.save(code)
    return code.code


class AuthorizationCodeGrant(_AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)

    def query_authorization_code(self, code, client):
        item = Backend.instance.query(
            models.AuthorizationCode, code=code, client=client
        )
        if item and not item[0].is_expired():
            return item[0]

    def delete_authorization_code(self, authorization_code):
        Backend.instance.delete(authorization_code)

    def authenticate_user(self, authorization_code):
        if authorization_code.subject and not authorization_code.subject.locked:
            return authorization_code.subject


class OpenIDCode(_OpenIDCode):
    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def get_jwt_config(self, grant):
        return get_jwt_config(grant)

    def generate_user_info(self, user, scope):
        return generate_user_info(user, scope)

    def get_audiences(self, request):
        client = request.client
        return [aud.client_id for aud in client.audience]


class PasswordGrant(_ResourceOwnerPasswordCredentialsGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def authenticate_user(self, username, password):
        user = Backend.instance.get_user_from_login(username)
        if not user:
            return None

        success, _ = Backend.instance.check_user_password(user, password)
        if not success:
            return None

        return user


class RefreshTokenGrant(_RefreshTokenGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def authenticate_refresh_token(self, refresh_token):
        token = Backend.instance.query(models.Token, refresh_token=refresh_token)
        if token and token[0].is_refresh_token_active():
            return token[0]

    def authenticate_user(self, credential):
        if credential.subject and not credential.subject.locked:
            return credential.subject

    def revoke_old_credential(self, credential):
        credential.revokation_date = datetime.datetime.now(datetime.timezone.utc)
        Backend.instance.save(credential)


class OpenIDImplicitGrant(_OpenIDImplicitGrant):
    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def get_jwt_config(self, grant=None):
        return get_jwt_config(grant)

    def generate_user_info(self, user, scope):
        return generate_user_info(user, scope)

    def get_audiences(self, request):
        client = request.client
        return [aud.client_id for aud in client.audience]


class OpenIDHybridGrant(_OpenIDHybridGrant):
    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)

    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def get_jwt_config(self, grant=None):
        return get_jwt_config(grant)

    def generate_user_info(self, user, scope):
        return generate_user_info(user, scope)

    def get_audiences(self, request):
        client = request.client
        return [aud.client_id for aud in client.audience]


def query_client(client_id):
    return Backend.instance.get(models.Client, client_id=client_id)


def save_token(token, request):
    now = datetime.datetime.now(datetime.timezone.utc)
    t = models.Token(
        token_id=gen_salt(48),
        type=token["token_type"],
        access_token=token["access_token"],
        issue_date=now,
        lifetime=token["expires_in"],
        scope=token["scope"].split(" "),
        client=request.client,
        refresh_token=token.get("refresh_token"),
        subject=request.user,
        audience=request.client.audience,
    )
    Backend.instance.save(t)


class BearerTokenValidator(_BearerTokenValidator):
    def authenticate_token(self, token_string):
        return Backend.instance.get(models.Token, access_token=token_string)


def query_token(token, token_type_hint):
    if token_type_hint == "access_token":
        return Backend.instance.get(models.Token, access_token=token)
    elif token_type_hint == "refresh_token":
        return Backend.instance.get(models.Token, refresh_token=token)

    item = Backend.instance.get(models.Token, access_token=token)
    if item:
        return item

    item = Backend.instance.get(models.Token, refresh_token=token)
    if item:
        return item

    return None


class RevocationEndpoint(_RevocationEndpoint):
    def query_token(self, token, token_type_hint):
        return query_token(token, token_type_hint)

    def revoke_token(self, token, request):
        token.revokation_date = datetime.datetime.now(datetime.timezone.utc)
        Backend.instance.save(token)


class IntrospectionEndpoint(_IntrospectionEndpoint):
    def query_token(self, token, token_type_hint):
        return query_token(token, token_type_hint)

    def check_permission(self, token, client, request):
        return client in token.audience

    def introspect_token(self, token):
        audience = [aud.client_id for aud in token.audience]
        return {
            "active": True,
            "client_id": token.client.client_id,
            "token_type": token.type,
            "username": token.subject.formatted_name,
            "scope": token.get_scope(),
            "sub": token.subject.user_name,
            "aud": audience,
            "iss": get_issuer(),
            "exp": token.get_expires_at(),
            "iat": token.get_issued_at(),
        }


class ClientManagementMixin:
    def authenticate_token(self, request):
        if current_app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_OPEN"]:
            return True

        auth_header = request.headers.get("Authorization")
        if not auth_header or not auth_header.lower().startswith("bearer "):
            return None

        bearer_token = auth_header.split()[1]
        if bearer_token not in (
            current_app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"]
            or []
        ):
            return None

        return True

    def get_server_metadata(self):
        result = openid_configuration()
        return result

    def resolve_public_key(self, request):
        # At the moment the only keypair accepted in software statement
        # is the one used to issues JWTs. This might change somedays.
        return current_app.config["CANAILLE_OIDC"]["JWT"]["PUBLIC_KEY"]

    def client_convert_data(self, **kwargs):
        if "client_id_issued_at" in kwargs:
            kwargs["client_id_issued_at"] = datetime.datetime.fromtimestamp(
                kwargs["client_id_issued_at"], datetime.timezone.utc
            )

        if "client_secret_expires_at" in kwargs:
            kwargs["client_secret_expires_at"] = datetime.datetime.fromtimestamp(
                kwargs["client_secret_expires_at"], datetime.timezone.utc
            )

        kwargs["scope"] = kwargs["scope"].split(" ")

        return kwargs


class ClientRegistrationEndpoint(ClientManagementMixin, _ClientRegistrationEndpoint):
    software_statement_alg_values_supported = ["RS256"]

    def generate_client_registration_info(self, client, request):
        return {"scope": " ".join(client.scope)}

    def save_client(self, client_info, client_metadata, request):
        if "scope" not in client_metadata:
            client_metadata["scope"] = "openid"
        client = models.Client(
            # this won't be needed when OIDC RP Initiated Logout is
            # directly implemented in authlib:
            # https://gitlab.com/yaal/canaille/-/issues/157
            post_logout_redirect_uris=request.data.get("post_logout_redirect_uris"),
            **self.client_convert_data(**client_info, **client_metadata),
        )
        Backend.instance.save(client)
        client.audience = [client]
        Backend.instance.save(client)
        return client


class ClientConfigurationEndpoint(ClientManagementMixin, _ClientConfigurationEndpoint):
    def authenticate_client(self, request):
        client_id = request.uri.split("/")[-1]
        return Backend.instance.get(models.Client, client_id=client_id)

    def revoke_access_token(self, request, token):
        pass

    def check_permission(self, client, request):
        return True

    def delete_client(self, client, request):
        Backend.instance.delete(client)

    def update_client(self, client, client_metadata, request):
        Backend.instance.update(client, **self.client_convert_data(**client_metadata))
        Backend.instance.save(client)
        return client

    def generate_client_registration_info(self, client, request):
        access_token = request.headers["Authorization"].split(" ")[1]
        return {
            "registration_client_uri": request.uri,
            "registration_access_token": access_token,
        }


class CodeChallenge(_CodeChallenge):
    def get_authorization_code_challenge(self, authorization_code):
        return authorization_code.challenge

    def get_authorization_code_challenge_method(self, authorization_code):
        return authorization_code.challenge_method


authorization = AuthorizationServer()
require_oauth = ResourceProtector()


def generate_access_token(client, grant_type, user, scope):
    if grant_type == "client_credentials":
        # Canaille could generate a JWT with iss/sub/aud/exp/iat/jti/scope/client_id
        # instead of a random string
        return gen_salt(48)

    audience = [client.client_id for client in client.audience]
    bearer_token_generator = authorization._token_generators["default"]
    kwargs = {
        "token": {},
        "user_info": generate_user_info(user, scope),
        "aud": audience,
        **get_jwt_config(grant_type),
    }
    kwargs["exp"] = bearer_token_generator._get_expires_in(client, grant_type)
    return generate_id_token(**kwargs)


def setup_oauth(app):
    app.config["OAUTH2_REFRESH_TOKEN_GENERATOR"] = True
    app.config["OAUTH2_ACCESS_TOKEN_GENERATOR"] = (
        "canaille.oidc.oauth.generate_access_token"
    )

    # hacky, but needed for tests as somehow the same 'authorization' object is used
    # between tests
    authorization.__init__()
    authorization.init_app(app, query_client=query_client, save_token=save_token)

    authorization.register_grant(PasswordGrant)
    authorization.register_grant(ImplicitGrant)
    authorization.register_grant(RefreshTokenGrant)
    authorization.register_grant(ClientCredentialsGrant)

    authorization.register_grant(
        AuthorizationCodeGrant,
        [
            OpenIDCode(require_nonce=app.config["CANAILLE_OIDC"]["REQUIRE_NONCE"]),
            CodeChallenge(required=True),
        ],
    )
    authorization.register_grant(OpenIDImplicitGrant)
    authorization.register_grant(OpenIDHybridGrant)

    require_oauth.register_token_validator(BearerTokenValidator())

    authorization.register_endpoint(IntrospectionEndpoint)
    authorization.register_endpoint(RevocationEndpoint)
    authorization.register_endpoint(ClientRegistrationEndpoint)
    authorization.register_endpoint(ClientConfigurationEndpoint)



import datetime
import logging


def test_signin_and_out(testclient, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "user" == session.get("attempt_login")

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()

    assert (
        "success",
        "Connection successful. Welcome John (johnny) Doe",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Succeed login attempt for user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert [user.id] == session.get("user_id")
        assert "attempt_login" not in session

    res = testclient.get("/login", status=302)

    res = testclient.get("/logout")
    assert (
        "success",
        "You have been disconnected. See you next time John (johnny) Doe",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Logout user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)


def test_visitor_logout(testclient, user):
    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/logout")
    res = res.follow(status=302)
    res = res.follow(status=200)
    assert (
        "success",
        "You have been disconnected. See you next time user",
    ) not in res.flashes

    with testclient.session_transaction() as session:
        assert not session.get("user_id")


def test_signin_wrong_password(testclient, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)
    res.form["password"] = "incorrect horse"
    res = res.form.submit(status=200)
    assert ("error", "Login failed, please check your information") in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed login attempt for user",
    ) in caplog.record_tuples


def test_signin_password_substring(testclient, user):
    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)
    res.form["password"] = "c"
    res = res.form.submit(status=200)
    assert ("error", "Login failed, please check your information") in res.flashes


def test_signin_bad_csrf(testclient, user):
    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    res = testclient.get("/login", status=200)

    res.form["login"] = "John (johnny) Doe"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = ""
    res = res.form.submit(status=200)
    assert ("error", "Login failed, please check your information") in res.flashes


def test_signin_with_alternate_attribute(testclient, user):
    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()
    res = res.follow(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert [user.id] == session.get("user_id")


def test_password_page_without_signin_in_redirects_to_login_page(testclient, user):
    res = testclient.get("/password", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with")
    ]


def test_password_page_already_logged_in(testclient, logged_user):
    res = testclient.get("/password", status=302)
    assert res.location == "/profile/user"


def test_wrong_login(testclient, user):
    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = True

    res = testclient.get("/login", status=200)
    res.form["login"] = "invalid"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "incorrect horse"
    res = res.form.submit(status=200)
    res.mustcontain(no="The login &#39;invalid&#39; does not exist")

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = False

    res = testclient.get("/login", status=200)
    res.form["login"] = "invalid"
    res = res.form.submit(status=200)
    res.mustcontain("The login &#39;invalid&#39; does not exist")


def test_signin_locked_account(testclient, user, backend):
    with testclient.session_transaction() as session:
        assert not session.get("user_id")

    user.lock_date = datetime.datetime.now(datetime.timezone.utc)
    backend.save(user)

    res = testclient.get("/login", status=200)
    res.form["login"] = "user"

    res = res.form.submit(status=302).follow(status=200)
    res.form["password"] = "correct horse battery staple"

    res = res.form.submit()
    res.mustcontain("Your account has been locked.")

    user.lock_date = None
    backend.save(user)



import ldap.filter
from flask import current_app

import canaille.core.models
import canaille.oidc.models

from .backend import LDAPBackend
from .ldapobject import LDAPObject


class User(canaille.core.models.User, LDAPObject):
    attribute_map = {
        "id": "entryUUID",
        "created": "createTimestamp",
        "last_modified": "modifyTimestamp",
        "user_name": "uid",
        "password": "userPassword",
        "preferred_language": "preferredLanguage",
        "family_name": "sn",
        "given_name": "givenName",
        "formatted_name": "cn",
        "display_name": "displayName",
        "emails": "mail",
        "phone_numbers": "telephoneNumber",
        "formatted_address": "postalAddress",
        "street": "street",
        "postal_code": "postalCode",
        "locality": "l",
        "region": "st",
        "photo": "jpegPhoto",
        "profile_url": "labeledURI",
        "employee_number": "employeeNumber",
        "department": "departmentNumber",
        "title": "title",
        "organization": "o",
        "groups": "memberOf",
        "lock_date": "pwdEndTime",
        "secret_token": "oathSecret",
        "last_otp_login": "oathLastLogin",
        "hotp_counter": "oathHOTPCounter",
        "one_time_password": "oathTokenPIN",
        "one_time_password_emission_date": "oathSecretTime",
        "password_failure_timestamps": "pwdFailureTime",
        "password_last_update": "pwdChangedTime",
    }

    def match_filter(self, filter):
        if isinstance(filter, str):
            conn = LDAPBackend.instance.connection
            return self.dn and conn.search_s(self.dn, ldap.SCOPE_SUBTREE, filter)

        return super().match_filter(filter)

    def save(self):
        if current_app.features.has_otp and not self.secret_token:
            self.initialize_otp()

        group_attr = self.python_attribute_to_ldap("groups")
        if group_attr not in self.changes:
            return

        # The LDAP attribute memberOf cannot directly be edited,
        # so this is needed to update the Group.member attribute
        # instead.
        old_groups = self.get_ldap_attribute(group_attr, lookup_changes=False) or []
        new_groups = self.get_ldap_attribute(group_attr, lookup_state=False) or []
        to_add = set(new_groups) - set(old_groups)
        to_del = set(old_groups) - set(new_groups)
        del self.changes[group_attr]

        yield

        for group in to_add:
            group.members = group.members + [self]
            LDAPBackend.instance.save(group)

        for group in to_del:
            # LDAP groups cannot be empty because groupOfNames.member
            # is a MUST attribute.
            # https://www.rfc-editor.org/rfc/rfc2256.html#section-7.10
            # TODO: properly manage the situation where one wants to
            # remove the last member of a group
            group.members = [member for member in group.members if member != self]
            LDAPBackend.instance.save(group)

        self.state[group_attr] = new_groups


class Group(canaille.core.models.Group, LDAPObject):
    attribute_map = {
        "id": "entryUUID",
        "created": "createTimestamp",
        "last_modified": "modifyTimestamp",
        "display_name": "cn",
        "members": "member",
        "description": "description",
    }


class Client(canaille.oidc.models.Client, LDAPObject):
    ldap_object_class = ["oauthClient"]
    base = "ou=clients,ou=oauth"
    rdn_attribute = "oauthClientID"

    client_info_attributes = {
        "client_id": "oauthClientID",
        "client_secret": "oauthClientSecret",
        "client_id_issued_at": "oauthIssueDate",
        "client_secret_expires_at": "oauthClientSecretExpDate",
    }

    client_metadata_attributes = {
        "client_name": "oauthClientName",
        "contacts": "oauthClientContact",
        "client_uri": "oauthClientURI",
        "redirect_uris": "oauthRedirectURIs",
        "logo_uri": "oauthLogoURI",
        "grant_types": "oauthGrantType",
        "response_types": "oauthResponseType",
        "scope": "oauthScope",
        "tos_uri": "oauthTermsOfServiceURI",
        "policy_uri": "oauthPolicyURI",
        "jwks_uri": "oauthJWKURI",
        "jwk": "oauthJWK",
        "token_endpoint_auth_method": "oauthTokenEndpointAuthMethod",
        "software_id": "oauthSoftwareID",
        "software_version": "oauthSoftwareVersion",
    }

    attribute_map = {
        "id": "entryUUID",
        "created": "createTimestamp",
        "last_modified": "modifyTimestamp",
        "preconsent": "oauthPreconsent",
        # post_logout_redirect_uris is not yet supported by authlib
        "post_logout_redirect_uris": "oauthPostLogoutRedirectURI",
        "audience": "oauthAudience",
        **client_info_attributes,
        **client_metadata_attributes,
    }


class AuthorizationCode(canaille.oidc.models.AuthorizationCode, LDAPObject):
    ldap_object_class = ["oauthAuthorizationCode"]
    base = "ou=authorizations,ou=oauth"
    rdn_attribute = "oauthAuthorizationCodeID"
    attribute_map = {
        "id": "entryUUID",
        "created": "createTimestamp",
        "last_modified": "modifyTimestamp",
        "authorization_code_id": "oauthAuthorizationCodeID",
        "code": "oauthCode",
        "client": "oauthClient",
        "subject": "oauthSubject",
        "redirect_uri": "oauthRedirectURI",
        "response_type": "oauthResponseType",
        "scope": "oauthScope",
        "nonce": "oauthNonce",
        "issue_date": "oauthAuthorizationDate",
        "lifetime": "oauthAuthorizationLifetime",
        "challenge": "oauthCodeChallenge",
        "challenge_method": "oauthCodeChallengeMethod",
        "revokation_date": "oauthRevokationDate",
    }


class Token(canaille.oidc.models.Token, LDAPObject):
    ldap_object_class = ["oauthToken"]
    base = "ou=tokens,ou=oauth"
    rdn_attribute = "oauthTokenID"
    attribute_map = {
        "id": "entryUUID",
        "created": "createTimestamp",
        "last_modified": "modifyTimestamp",
        "token_id": "oauthTokenID",
        "access_token": "oauthAccessToken",
        "client": "oauthClient",
        "subject": "oauthSubject",
        "type": "oauthTokenType",
        "refresh_token": "oauthRefreshToken",
        "scope": "oauthScope",
        "issue_date": "oauthIssueDate",
        "lifetime": "oauthTokenLifetime",
        "revokation_date": "oauthRevokationDate",
        "audience": "oauthAudience",
    }


class Consent(canaille.oidc.models.Consent, LDAPObject):
    ldap_object_class = ["oauthConsent"]
    base = "ou=consents,ou=oauth"
    rdn_attribute = "cn"
    attribute_map = {
        "id": "entryUUID",
        "created": "createTimestamp",
        "last_modified": "modifyTimestamp",
        "consent_id": "cn",
        "subject": "oauthSubject",
        "client": "oauthClient",
        "scope": "oauthScope",
        "issue_date": "oauthIssueDate",
        "revokation_date": "oauthRevokationDate",
    }



"""initial migration.

Represents the state of the database in version 0.0.56

Revision ID: 1736443094
Revises:
Create Date: 2025-01-09 18:18:14.276914

"""

from collections.abc import Sequence

import sqlalchemy as sa
import sqlalchemy_utils.types.password
from alembic import op

import canaille.backends.sql.utils

# revision identifiers, used by Alembic.
revision: str = "1736443094"
down_revision: str | None = None
branch_labels: str | Sequence[str] | None = ("default",)
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "client",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "last_modified",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("preconsent", sa.Boolean(), nullable=True),
        sa.Column("post_logout_redirect_uris", sa.JSON(), nullable=True),
        sa.Column("client_id", sa.String(), nullable=True),
        sa.Column("client_secret", sa.String(), nullable=True),
        sa.Column(
            "client_id_issued_at",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "client_secret_expires_at",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("client_name", sa.String(), nullable=True),
        sa.Column("contacts", sa.JSON(), nullable=True),
        sa.Column("client_uri", sa.String(), nullable=True),
        sa.Column("redirect_uris", sa.JSON(), nullable=True),
        sa.Column("logo_uri", sa.String(), nullable=True),
        sa.Column("grant_types", sa.JSON(), nullable=True),
        sa.Column("response_types", sa.JSON(), nullable=True),
        sa.Column("scope", sa.JSON(), nullable=True),
        sa.Column("tos_uri", sa.String(), nullable=True),
        sa.Column("policy_uri", sa.String(), nullable=True),
        sa.Column("jwks_uri", sa.String(), nullable=True),
        sa.Column("jwk", sa.String(), nullable=True),
        sa.Column("token_endpoint_auth_method", sa.String(), nullable=True),
        sa.Column("software_id", sa.String(), nullable=True),
        sa.Column("software_version", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "group",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "last_modified",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("display_name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "user",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "last_modified",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("user_name", sa.String(), nullable=False),
        sa.Column(
            "password",
            sqlalchemy_utils.types.password.PasswordType(max_length=4096),
            nullable=True,
        ),
        sa.Column("preferred_language", sa.String(), nullable=True),
        sa.Column("family_name", sa.String(), nullable=True),
        sa.Column("given_name", sa.String(), nullable=True),
        sa.Column("formatted_name", sa.String(), nullable=True),
        sa.Column("display_name", sa.String(), nullable=True),
        sa.Column("emails", sa.JSON(), nullable=True),
        sa.Column("phone_numbers", sa.JSON(), nullable=True),
        sa.Column("formatted_address", sa.String(), nullable=True),
        sa.Column("street", sa.String(), nullable=True),
        sa.Column("postal_code", sa.String(), nullable=True),
        sa.Column("locality", sa.String(), nullable=True),
        sa.Column("region", sa.String(), nullable=True),
        sa.Column("photo", sa.LargeBinary(), nullable=True),
        sa.Column("profile_url", sa.String(), nullable=True),
        sa.Column("employee_number", sa.String(), nullable=True),
        sa.Column("department", sa.String(), nullable=True),
        sa.Column("title", sa.String(), nullable=True),
        sa.Column("organization", sa.String(), nullable=True),
        sa.Column(
            "lock_date",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_name"),
    )
    op.create_table(
        "authorization_code",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "last_modified",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("authorization_code_id", sa.String(), nullable=True),
        sa.Column("code", sa.String(), nullable=True),
        sa.Column("client_id", sa.String(), nullable=False),
        sa.Column("subject_id", sa.String(), nullable=False),
        sa.Column("redirect_uri", sa.String(), nullable=True),
        sa.Column("response_type", sa.String(), nullable=True),
        sa.Column("scope", sa.JSON(), nullable=True),
        sa.Column("nonce", sa.String(), nullable=True),
        sa.Column(
            "issue_date",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("lifetime", sa.Integer(), nullable=True),
        sa.Column("challenge", sa.String(), nullable=True),
        sa.Column("challenge_method", sa.String(), nullable=True),
        sa.Column(
            "revokation_date",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["subject_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "client_audience_association_table",
        sa.Column("audience_id", sa.String(), nullable=True),
        sa.Column("client_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["audience_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.PrimaryKeyConstraint("audience_id", "client_id"),
    )
    op.create_table(
        "consent",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "last_modified",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("consent_id", sa.String(), nullable=True),
        sa.Column("subject_id", sa.String(), nullable=False),
        sa.Column("client_id", sa.String(), nullable=False),
        sa.Column("scope", sa.JSON(), nullable=True),
        sa.Column(
            "issue_date",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "revokation_date",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["subject_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "membership_association_table",
        sa.Column("user_id", sa.String(), nullable=False),
        sa.Column("group_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["group_id"],
            ["group.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("user_id", "group_id"),
    )
    op.create_table(
        "token",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column(
            "last_modified",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("token_id", sa.String(), nullable=True),
        sa.Column("access_token", sa.String(), nullable=True),
        sa.Column("client_id", sa.String(), nullable=False),
        sa.Column("subject_id", sa.String(), nullable=False),
        sa.Column("type", sa.String(), nullable=True),
        sa.Column("refresh_token", sa.String(), nullable=True),
        sa.Column("scope", sa.JSON(), nullable=True),
        sa.Column(
            "issue_date",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.Column("lifetime", sa.Integer(), nullable=True),
        sa.Column(
            "revokation_date",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["subject_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "token_audience_association_table",
        sa.Column("token_id", sa.String(), nullable=True),
        sa.Column("client_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["token_id"],
            ["token.id"],
        ),
        sa.PrimaryKeyConstraint("token_id", "client_id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("token_audience_association_table")
    op.drop_table("token")
    op.drop_table("membership_association_table")
    op.drop_table("consent")
    op.drop_table("client_audience_association_table")
    op.drop_table("authorization_code")
    op.drop_table("user")
    op.drop_table("group")
    op.drop_table("client")
    # ### end Alembic commands ###



import json
import logging
from unittest import mock

import pytest

from canaille.commands import cli


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_reset_otp_by_id(testclient, backend, caplog, user_otp, otp_method):
    """Reset one-time password authentication for a user by its id."""
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    old_token = user_otp.secret_token
    assert old_token is not None
    assert user_otp.last_otp_login is not None

    runner = testclient.app.test_cli_runner()
    res = runner.invoke(
        cli,
        [
            "reset-otp",
            user_otp.id,
        ],
    )
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1235, somewhere",
        "formatted_name": "John (johnny) Doe",
        "given_name": "John",
        "id": user_otp.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
        "hotp_counter": 1,
        "secret_token": mock.ANY,
    }
    assert (
        "canaille",
        logging.SECURITY,
        "Reset one-time password authentication from CLI for user",
    ) in caplog.record_tuples
    backend.reload(user_otp)
    assert user_otp.secret_token is not None
    assert user_otp.secret_token != old_token
    assert user_otp.last_otp_login is None
    if otp_method == "HOTP":
        assert user_otp.hotp_counter == 1


def test_reset_otp_unknown_id(testclient):
    """Error case for trying to reset one-time password authentication for an invalid user."""
    runner = testclient.app.test_cli_runner()
    res = runner.invoke(
        cli,
        [
            "reset-otp",
            "invalid",
        ],
    )
    assert res.exit_code == 1, res.stdout
    assert res.stdout == "Error: No user with id 'invalid'\n"



import datetime

from flask import Blueprint
from flask import abort
from flask import current_app
from flask import flash
from flask import redirect
from flask import request
from flask import session
from flask import url_for

from canaille.app import build_hash
from canaille.app import get_b64encoded_qr_image
from canaille.app import mask_email
from canaille.app import mask_phone
from canaille.app.flask import smtp_needed
from canaille.app.i18n import gettext as _
from canaille.app.session import current_user
from canaille.app.session import login_user
from canaille.app.session import logout_user
from canaille.app.templating import render_template
from canaille.backends import Backend
from canaille.core.endpoints.forms import TwoFactorForm
from canaille.core.models import SEND_NEW_OTP_DELAY

from ..mails import send_password_initialization_mail
from ..mails import send_password_reset_mail
from .forms import FirstLoginForm
from .forms import ForgottenPasswordForm
from .forms import LoginForm
from .forms import PasswordForm
from .forms import PasswordResetForm

bp = Blueprint("auth", __name__)


@bp.context_processor
def global_processor():
    return {
        "menu": False,
    }


@bp.route("/login", methods=("GET", "POST"))
def login():
    if current_user():
        return redirect(
            url_for("core.account.profile_edition", edited_user=current_user())
        )

    form = LoginForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"
    form["login"].render_kw["placeholder"] = Backend.instance.login_placeholder()

    if not request.form or form.form_control():
        return render_template("core/login.html", form=form)

    user = Backend.instance.get_user_from_login(form.login.data)
    if user and not user.has_password() and current_app.features.has_smtp:
        return redirect(url_for("core.auth.firstlogin", user=user))

    if not form.validate():
        logout_user()
        flash(_("Login failed, please check your information"), "error")
        return render_template("core/login.html", form=form)

    session["attempt_login"] = form.login.data
    return redirect(url_for("core.auth.password"))


@bp.route("/password", methods=("GET", "POST"))
def password():
    if current_user():
        return redirect(
            url_for("core.account.profile_edition", edited_user=current_user())
        )

    if "attempt_login" not in session:
        flash(_("Cannot remember the login you attempted to sign in with"), "warning")
        return redirect(url_for("core.auth.login"))

    form = PasswordForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"

    if not request.form or form.form_control():
        return render_template(
            "core/password.html", form=form, username=session["attempt_login"]
        )

    user = Backend.instance.get_user_from_login(session["attempt_login"])
    if user and not user.has_password() and current_app.features.has_smtp:
        return redirect(url_for("core.auth.firstlogin", user=user))

    if not form.validate() or not user:
        logout_user()
        flash(_("Login failed, please check your information"), "error")
        return render_template(
            "core/password.html", form=form, username=session["attempt_login"]
        )

    success, message = Backend.instance.check_user_password(user, form.password.data)
    if not success:
        logout_user()
        current_app.logger.security(
            f"Failed login attempt for {session['attempt_login']}"
        )
        flash(message or _("Login failed, please check your information"), "error")
        return render_template(
            "core/password.html", form=form, username=session["attempt_login"]
        )

    otp_methods = []
    if current_app.features.has_otp:
        otp_methods.append(current_app.features.otp_method)  # TOTP or HOTP
    if current_app.features.has_email_otp:
        otp_methods.append("EMAIL_OTP")
    if current_app.features.has_sms_otp:
        otp_methods.append("SMS_OTP")

    if otp_methods:
        session["remaining_otp_methods"] = otp_methods
        session["attempt_login_with_correct_password"] = session.pop("attempt_login")
        return redirect_to_verify_mfa(
            user, otp_methods[0], url_for("core.auth.password")
        )
    else:
        current_app.logger.security(
            f"Succeed login attempt for {session['attempt_login']}"
        )
        del session["attempt_login"]
        login_user(user)
        flash(
            _("Connection successful. Welcome %(user)s", user=user.formatted_name),
            "success",
        )
        return redirect(
            session.pop("redirect-after-login", url_for("core.account.index"))
        )


@bp.route("/logout")
def logout():
    user = current_user()

    if user:
        current_app.logger.security(f"Logout {user.identifier}")

        flash(
            _(
                "You have been disconnected. See you next time %(user)s",
                user=user.formatted_name,
            ),
            "success",
        )
        logout_user()
    return redirect("/")


@bp.route("/firstlogin/<user:user>", methods=("GET", "POST"))
def firstlogin(user):
    if user.has_password():
        abort(404)

    form = FirstLoginForm(request.form or None)
    if not request.form:
        return render_template("core/firstlogin.html", form=form, user=user)

    form.validate()

    statuses = [send_password_initialization_mail(user, email) for email in user.emails]
    success = all(statuses)
    if success:
        flash(
            _(
                "A password initialization link has been sent at your email address. "
                "You should receive it within a few minutes."
            ),
            "success",
        )
    else:
        flash(_("Could not send the password initialization email"), "error")

    return render_template("core/firstlogin.html", form=form)


@bp.route("/reset", methods=["GET", "POST"])
@smtp_needed()
def forgotten():
    if not current_app.config["CANAILLE"]["ENABLE_PASSWORD_RECOVERY"]:
        abort(404)

    form = ForgottenPasswordForm(request.form)
    if not request.form:
        return render_template("core/forgotten-password.html", form=form)

    if not form.validate():
        flash(_("Could not send the password reset link."), "error")
        return render_template("core/forgotten-password.html", form=form)

    user = Backend.instance.get_user_from_login(form.login.data)
    success_message = _(
        "A password reset link has been sent at your email address. "
        "You should receive it within a few minutes."
    )
    if current_app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] and (
        not user or not user.can_edit_self
    ):
        flash(success_message, "success")
        return render_template("core/forgotten-password.html", form=form)

    if not user.can_edit_self:
        flash(
            _(
                "The user '%(user)s' does not have permissions to update their password. "
                "We cannot send a password reset email.",
                user=user.formatted_name,
            ),
            "error",
        )
        return render_template("core/forgotten-password.html", form=form)

    success = True
    for email in user.emails:
        if not send_password_reset_mail(user, email):
            success = False
        current_app.logger.security(
            f"Sending a reset password mail to {email} for {user.user_name}"
        )

    if success:
        flash(success_message, "success")
    else:
        flash(
            _("We encountered an issue while we sent the password recovery email."),
            "error",
        )

    return render_template("core/forgotten-password.html", form=form)


@bp.route("/reset/<user:user>/<hash>", methods=["GET", "POST"])
def reset(user, hash):
    if not current_app.config["CANAILLE"]["ENABLE_PASSWORD_RECOVERY"]:
        abort(404)

    form = PasswordResetForm(request.form)
    hashes = {
        build_hash(
            user.identifier,
            email,
            user.password if user.has_password() else "",
        )
        for email in user.emails
    }
    if not user or hash not in hashes:
        flash(
            _("The password reset link that brought you here was invalid."),
            "error",
        )
        return redirect(url_for("core.account.index"))

    if request.form and form.validate():
        Backend.instance.set_user_password(user, form.password.data)
        login_user(user)

        flash(_("Your password has been updated successfully"), "success")
        return redirect(
            session.pop(
                "redirect-after-login",
                url_for("core.account.profile_edition", edited_user=user),
            )
        )

    return render_template("core/reset-password.html", form=form, user=user, hash=hash)


@bp.route("/setup-mfa")
def setup_two_factor_auth():
    if not current_app.features.has_otp:
        abort(404)

    if current_user():
        return redirect(
            url_for("core.account.profile_edition", edited_user=current_user())
        )

    if "attempt_login_with_correct_password" not in session:
        flash(_("Cannot remember the login you attempted to sign in with"), "warning")
        return redirect(url_for("core.auth.login"))

    user = Backend.instance.get_user_from_login(
        session["attempt_login_with_correct_password"]
    )

    uri = user.get_otp_authentication_setup_uri()
    base64_qr_image = get_b64encoded_qr_image(uri)
    return render_template(
        "core/setup-mfa.html",
        secret=user.secret_token,
        qr_image=base64_qr_image,
        user=user,
    )


@bp.route("/verify-mfa", methods=["GET", "POST"])
def verify_two_factor_auth():
    if current_user():
        return redirect(
            url_for("core.account.profile_edition", edited_user=current_user())
        )

    if (
        "attempt_login_with_correct_password" not in session
        or "remaining_otp_methods" not in session
        or not session["remaining_otp_methods"]
    ):
        flash(_("Cannot remember the login you attempted to sign in with"), "warning")
        return redirect(url_for("core.auth.login"))

    current_otp_method = session["remaining_otp_methods"][0]
    if (
        (current_otp_method in ["TOTP", "HOTP"] and not current_app.features.has_otp)
        or (
            current_otp_method == "EMAIL_OTP" and not current_app.features.has_email_otp
        )
        or (current_otp_method == "SMS_OTP" and not current_app.features.has_sms_otp)
    ):
        abort(404)

    form = TwoFactorForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"

    if not request.form or form.form_control():
        return render_template(
            "core/verify-mfa.html",
            form=form,
            username=session["attempt_login_with_correct_password"],
            method=current_otp_method,
        )

    user = Backend.instance.get_user_from_login(
        session["attempt_login_with_correct_password"]
    )

    if form.validate() and user.is_otp_valid(form.otp.data, current_otp_method):
        session["remaining_otp_methods"].pop(0)
        if session["remaining_otp_methods"]:
            return redirect_to_verify_mfa(
                user,
                session["remaining_otp_methods"][0],
                url_for("core.auth.verify_two_factor_auth"),
            )
        else:
            user.last_otp_login = datetime.datetime.now(datetime.timezone.utc)
            Backend.instance.save(user)
            current_app.logger.security(
                f"Succeed login attempt for {session['attempt_login_with_correct_password']}"
            )
            del session["attempt_login_with_correct_password"]
            login_user(user)
            flash(
                _(
                    "Connection successful. Welcome %(user)s",
                    user=user.formatted_name,
                ),
                "success",
            )
            return redirect(
                session.pop("redirect-after-login", url_for("core.account.index"))
            )
    else:
        flash(
            "The one-time password you entered is invalid. Please try again",
            "error",
        )
        current_app.logger.security(
            f"Failed login attempt (wrong OTP) for {session['attempt_login_with_correct_password']}"
        )
        return redirect(url_for("core.auth.verify_two_factor_auth"))


@bp.route("/send-mail-otp", methods=["POST"])
def send_mail_otp():
    if not current_app.features.has_email_otp:
        abort(404)

    if current_user():
        return redirect(
            url_for("core.account.profile_edition", edited_user=current_user())
        )

    if "attempt_login_with_correct_password" not in session:
        flash(_("Cannot remember the login you attempted to sign in with"), "warning")
        return redirect(url_for("core.auth.login"))

    user = Backend.instance.get_user_from_login(
        session["attempt_login_with_correct_password"]
    )

    if user.can_send_new_otp():
        if user.generate_and_send_otp_mail():
            Backend.instance.save(user)
            current_app.logger.security(
                f"Sent one-time password for {session['attempt_login_with_correct_password']} to {user.emails[0]}"
            )
            flash(
                "Code successfully sent!",
                "success",
            )
        else:
            flash("Error while sending one-time password. Please try again.", "danger")
    else:
        flash(
            f"Too many attempts. Please try again in {SEND_NEW_OTP_DELAY} seconds.",
            "danger",
        )

    return redirect(url_for("core.auth.verify_two_factor_auth"))


@bp.route("/send-sms-otp", methods=["POST"])
def send_sms_otp():
    if not current_app.features.has_sms_otp:
        abort(404)

    if current_user():
        return redirect(
            url_for("core.account.profile_edition", edited_user=current_user())
        )

    if "attempt_login_with_correct_password" not in session:
        flash(_("Cannot remember the login you attempted to sign in with"), "warning")
        return redirect(url_for("core.auth.login"))

    user = Backend.instance.get_user_from_login(
        session["attempt_login_with_correct_password"]
    )

    if user.can_send_new_otp():
        if user.generate_and_send_otp_sms():
            Backend.instance.save(user)
            current_app.logger.security(
                f"Sent one-time password for {session['attempt_login_with_correct_password']} to {user.phone_numbers[0]}"
            )
            flash(
                "Code successfully sent!",
                "success",
            )
        else:
            flash("Error while sending one-time password. Please try again.", "danger")
    else:
        flash(
            f"Too many attempts. Please try again in {SEND_NEW_OTP_DELAY} seconds.",
            "danger",
        )

    return redirect(url_for("core.auth.verify_two_factor_auth"))


def redirect_to_verify_mfa(user, otp_method, fail_redirect_url):
    if otp_method in ["HOTP", "TOTP"]:
        if not user.last_otp_login:
            flash(
                "You have not enabled multi-factor authentication. Please enable it first to login.",
                "info",
            )
            return redirect(url_for("core.auth.setup_two_factor_auth"))
        flash(
            "Please enter the one-time password from your authenticator app.",
            "info",
        )
        return redirect(url_for("core.auth.verify_two_factor_auth"))
    elif otp_method == "EMAIL_OTP":
        if user.can_send_new_otp():
            if user.generate_and_send_otp_mail():
                Backend.instance.save(user)
                flash(
                    f"A one-time password has been sent to your email address {mask_email(user.emails[0])}. Please enter it below to login.",
                    "info",
                )
                current_app.logger.security(
                    f"Sent one-time password for {session['attempt_login_with_correct_password']} to {user.emails[0]}"
                )
                return redirect(url_for("core.auth.verify_two_factor_auth"))
            else:
                flash(
                    "Error while sending one-time password. Please try again.", "danger"
                )
                return redirect(fail_redirect_url)
        else:
            flash(
                f"Too many attempts. Please try again in {SEND_NEW_OTP_DELAY} seconds.",
                "danger",
            )
            return redirect(fail_redirect_url)
    else:  # sms
        if user.can_send_new_otp():
            if user.generate_and_send_otp_sms():
                Backend.instance.save(user)
                flash(
                    f"A one-time password has been sent to your phone number {mask_phone(user.phone_numbers[0])}. Please enter it below to login.",
                    "info",
                )
                current_app.logger.security(
                    f"Sent one-time password for {session['attempt_login_with_correct_password']} to {user.phone_numbers[0]}"
                )
                return redirect(url_for("core.auth.verify_two_factor_auth"))
            else:
                flash(
                    "Error while sending one-time password. Please try again.",
                    "danger",
                )
                return redirect(fail_redirect_url)
        else:
            flash(
                f"Too many attempts. Please try again in {SEND_NEW_OTP_DELAY} seconds.",
                "danger",
            )
            return redirect(fail_redirect_url)
